"""Matplotlib stylesheets for plot styling"""
from pkg_resources import resource_filename

base_style = resource_filename("euphonic.styles", "base.mplstyle")
intensity_widget_style = resource_filename("euphonic.styles",
                                           "intensity_widget.mplstyle")
